from typing import Any, Optional, Mapping, Sequence

from kubragen.configfile import ConfigFile, ConfigFileOutput, ConfigFileOutput_Dict, ConfigFile_Extend, \
    ConfigFileExtensionData, ConfigFileExtension
from kubragen.helper import QuotedStr
from kubragen.merger import Merger
from kubragen.option import OptionDef
from kubragen.options import OptionGetter, Options, option_root_get


class PrometheusConfigFileOptions(Options):
    """
    Options for Prometheus config file.

    .. list-table::
        :header-rows: 1

        * - option
          - description
          - allowed types
          - default value
        * - config.extra_config
          - extra config to add to config file
          - Mapping
          - ```{}```
        * - scrape.prometheus.enabled
          - whether prometheus scrape is enabled
          - bool
          - ```True```
        * - scrape.prometheus.job_name
          - prometheus scrape job name
          - str
          - ```prometheus```
        * - scrape.prometheus.extra_config
          - extra config to add to prometheus job
          - Mapping
          - ```{}```
    """
    def define_options(self) -> Optional[Any]:
        """
        Declare the options for the Prometheus config file.

        :return: The supported options
        """
        return {
            'config': {
                'merge_config': OptionDef(default_value={}, allowed_types=[Mapping]),
            },
            'scrape': {
                'prometheus': {
                    'enabled': OptionDef(required=True, default_value=False, allowed_types=[bool]),
                    'job_name': OptionDef(default_value='prometheus', allowed_types=[str]),
                    'extra_config': OptionDef(default_value={}, allowed_types=[Mapping]),
                }
            },
        }


class PrometheusConfigFile(ConfigFile_Extend):
    """
    Prometheus main configuration file in YAML format.
    """
    options: PrometheusConfigFileOptions

    def __init__(self, options: Optional[PrometheusConfigFileOptions] = None,
                 extensions: Optional[Sequence[ConfigFileExtension]] = None):
        super().__init__(extensions)
        if options is None:
            options = PrometheusConfigFileOptions()
        self.options = options

    def option_get(self, name: str):
        return option_root_get(self.options, name)

    def init_value(self, options: OptionGetter) -> ConfigFileExtensionData:
        ret = ConfigFileExtensionData({})

        if self.option_get('scrape.prometheus.enabled'):
            Merger.merge(ret.data, {
                'scrape_configs': [Merger.merge({
                    'job_name': QuotedStr(self.option_get('scrape.prometheus.job_name')),
                    'static_configs': [{
                        'targets': [
                            'localhost:9090',
                        ]
                    }],
                }, self.option_get('scrape.prometheus.extra_config'))]
            })

        return ret

    def finish_value(self, options: OptionGetter, data: ConfigFileExtensionData) -> ConfigFileOutput:
        if self.option_get('config.merge_config') is not None:
            Merger.merge(data.data, self.option_get('config.merge_config'))
        return ConfigFileOutput_Dict(data.data)



from typing import List, Optional

from kubragen import KubraGen
from kubragen.builder import Builder
from kubragen.configfile import ConfigFileRenderMulti, ConfigFileRender_RawStr, ConfigFileRender_Yaml
from kubragen.data import ValueData
from kubragen.exception import InvalidParamError, InvalidNameError
from kubragen.helper import LiteralStr
from kubragen.kdatahelper import KDataHelper_Volume
from kubragen.object import ObjectItem, Object
from kubragen.types import TBuild, TBuildItem

from .configfile import PrometheusConfigFile
from .option import PrometheusOptions


class PrometheusBuilder(Builder):
    """
    Prometheus builder.

    Based on `prometheus-community/helm-charts <https://github.com/prometheus-community/helm-charts>`_.

    .. list-table::
        :header-rows: 1

        * - build
          - description
        * - BUILD_ACCESSCONTROL
          - creates service account, roles, and roles bindings
        * - BUILD_CONFIG
          - creates ConfigMap
        * - BUILD_SERVICE
          - creates StatefulSet and Services

    .. list-table::
        :header-rows: 1

        * - build item
          - description
        * - BUILDITEM_SERVICE_ACCOUNT
          - ServiceAccount
        * - BUILDBUILDITEM_CLUSTER_ROLEITEM_ROLE
          - ClusterRole
        * - BUILDITEM_CLUSTER_ROLE_BINDING
          - ClusterRoleBinding
        * - BUILDITEM_CONFIG
          - ConfigMap
        * - BUILDITEM_STATEFULSET
          - StatefulSet
        * - BUILDITEM_SERVICE
          - Service

    .. list-table::
        :header-rows: 1

        * - object name
          - description
          - default value
        * - config
          - ConfigMap
          - ```<basename>-config```
        * - service
          - Service
          - ```<basename>```
        * - service-account
          - ServiceAccount
          - ```<basename>```
        * - cluster-role
          - ClusterRole
          - ```<basename>```
        * - cluster-role-binding
          - ClusterRoleBinding
          - ```<basename>```
        * - statefulset
          - StatefulSet
          - ```<basename>```
        * - pod-label-all
          - label *app* to be used by selection
          - ```<basename>```
    """
    options: PrometheusOptions
    configfile: Optional[str]
    _namespace: str

    SOURCE_NAME = 'kg_prometheus'

    BUILD_ACCESSCONTROL: TBuild = 'accesscontrol'
    BUILD_CONFIG: TBuild = 'config'
    BUILD_SERVICE: TBuild = 'service'

    BUILDITEM_SERVICE_ACCOUNT: TBuildItem = 'service-account'
    BUILDITEM_CLUSTER_ROLE: TBuildItem = 'cluster-role'
    BUILDITEM_CLUSTER_ROLE_BINDING: TBuildItem = 'cluster-role-binding'
    BUILDITEM_CONFIG: TBuildItem = 'config'
    BUILDITEM_STATEFULSET: TBuildItem = 'statefulset'
    BUILDITEM_SERVICE: TBuildItem = 'service'

    def __init__(self, kubragen: KubraGen, options: Optional[PrometheusOptions] = None):
        super().__init__(kubragen)
        if options is None:
            options = PrometheusOptions()
        self.options = options
        self.configfile = None

        self._namespace = self.option_get('namespace')

        if self.option_get('config.authorization.serviceaccount_create') is not False:
            serviceaccount_name = self.basename()
        else:
            serviceaccount_name = self.option_get('config.authorization.serviceaccount_use')
            if serviceaccount_name == '':
                serviceaccount_name = None

        if self.option_get('config.authorization.roles_create') is not False:
            role_name = self.basename()
        else:
            role_name = None

        if self.option_get('config.authorization.roles_bind') is not False:
            if serviceaccount_name is None:
                raise InvalidParamError('To bind roles a service account is required')
            if role_name is None:
                raise InvalidParamError('To bind roles the role must be created')
            rolebinding_name = self.basename()
        else:
            rolebinding_name = None

        self.object_names_init({
            'config': self.basename('-config'),
            'service': self.basename(),
            'service-account': serviceaccount_name,
            'cluster-role': role_name,
            'cluster-role-binding': rolebinding_name,
            'statefulset': self.basename(),
            'pod-label-app': self.basename(),
        })

    def option_get(self, name: str):
        return self.kubragen.option_root_get(self.options, name)

    def configfile_get(self) -> str:
        if self.configfile is None:
            configfile = self.option_get('config.prometheus_config')
            if configfile is None:
                configfile = PrometheusConfigFile()
            if isinstance(configfile, str):
                self.configfile = configfile
            else:
                configfilerender = ConfigFileRenderMulti([
                    ConfigFileRender_Yaml(),
                    ConfigFileRender_RawStr()
                ])
                self.configfile = configfilerender.render(configfile.get_value(self))
        return self.configfile

    def basename(self, suffix: str = ''):
        return '{}{}'.format(self.option_get('basename'), suffix)

    def namespace(self):
        return self._namespace

    def build_names(self) -> List[TBuild]:
        return [self.BUILD_ACCESSCONTROL, self.BUILD_CONFIG, self.BUILD_SERVICE]

    def build_names_required(self) -> List[TBuild]:
        ret = [self.BUILD_CONFIG, self.BUILD_SERVICE]
        if self.option_get('config.authorization.serviceaccount_create') is not False or \
                self.option_get('config.authorization.roles_create') is not False:
            ret.append(self.BUILD_ACCESSCONTROL)
        return ret

    def builditem_names(self) -> List[TBuildItem]:
        return [
            self.BUILDITEM_SERVICE_ACCOUNT,
            self.BUILDITEM_CLUSTER_ROLE,
            self.BUILDITEM_CLUSTER_ROLE_BINDING,
            self.BUILDITEM_CONFIG,
            self.BUILDITEM_STATEFULSET,
            self.BUILDITEM_SERVICE,
        ]

    def internal_build(self, buildname: TBuild) -> List[ObjectItem]:
        if buildname == self.BUILD_ACCESSCONTROL:
            return self.internal_build_accesscontrol()
        elif buildname == self.BUILD_CONFIG:
            return self.internal_build_config()
        elif buildname == self.BUILD_SERVICE:
            return self.internal_build_service()
        else:
            raise InvalidNameError('Invalid build name: "{}"'.format(buildname))

    def internal_build_accesscontrol(self) -> List[ObjectItem]:
        ret = []

        if self.option_get('config.authorization.serviceaccount_create') is not False:
            ret.extend([
                Object({
                    'apiVersion': 'v1',
                    'kind': 'ServiceAccount',
                    'metadata': {
                        'name': self.object_name('service-account'),
                        'namespace': self.namespace(),
                    }
                }, name=self.BUILDITEM_SERVICE_ACCOUNT, source=self.SOURCE_NAME, instance=self.basename()),
            ])

        if self.option_get('config.authorization.roles_create') is not False:
            ret.extend([
                Object({
                    'apiVersion': 'rbac.authorization.k8s.io/v1',
                    'kind': 'ClusterRole',
                    'metadata': {
                        'name': self.object_name('cluster-role'),
                    },
                    'rules': [{
                        'apiGroups': [''],
                        'resources': [
                            'nodes',
                            'nodes/metrics',
                            'services',
                            'endpoints',
                            'pods'
                        ],
                        'verbs': ['get', 'list', 'watch']
                    },
                    {
                        'apiGroups': ['extensions'],
                        'resources': ['ingresses'],
                        'verbs': ['get', 'list', 'watch']
                    },
                    {
                        'nonResourceURLs': ['/metrics', '/metrics/cadvisor'],
                        'verbs': ['get']
                    }]
                }, name=self.BUILDITEM_CLUSTER_ROLE, source=self.SOURCE_NAME, instance=self.basename()),
            ])

        if self.option_get('config.authorization.roles_bind') is not False:
            ret.extend([
                Object({
                    'apiVersion': 'rbac.authorization.k8s.io/v1',
                    'kind': 'ClusterRoleBinding',
                    'metadata': {
                        'name': self.object_name('cluster-role-binding'),
                    },
                    'subjects': [{
                        'kind': 'ServiceAccount',
                        'name': self.object_name('service-account'),
                        'namespace': self.namespace(),
                    }],
                    'roleRef': {
                        'apiGroup': 'rbac.authorization.k8s.io',
                        'kind': 'ClusterRole',
                        'name': self.object_name('cluster-role'),
                    }
                }, name=self.BUILDITEM_CLUSTER_ROLE_BINDING, source=self.SOURCE_NAME, instance=self.basename())
            ])

        return ret

    def internal_build_config(self) -> List[ObjectItem]:
        ret = [
            Object({
                'apiVersion': 'v1',
                'kind': 'ConfigMap',
                'metadata': {
                    'name': self.object_name('config'),
                    'namespace': self.namespace(),
                },
                'data': {
                    'prometheus.yml': LiteralStr(self.configfile_get()),
                },
            }, name=self.BUILDITEM_CONFIG, source=self.SOURCE_NAME, instance=self.basename())
        ]
        return ret

    def internal_build_service(self) -> List[ObjectItem]:
        ret = [Object({
            'kind': 'Service',
            'apiVersion': 'v1',
            'metadata': {
                'name': self.object_name('service'),
                'namespace': self.namespace(),
            },
            'spec': {
                'selector': {
                    'app': self.object_name('pod-label-app')
                },
                'ports': [{
                    'protocol': 'TCP',
                    'port': self.option_get('config.service_port'),
                    'targetPort': 9090
                }]
            }
        }, name=self.BUILDITEM_SERVICE, source=self.SOURCE_NAME, instance=self.basename()), Object({
            'apiVersion': 'apps/v1',
            'kind': 'StatefulSet',
            'metadata': {
                'name': self.object_name('statefulset'),
                'namespace': self.namespace(),
                'labels': {
                    'app': self.object_name('pod-label-app'),
                },
            },
            'spec': {
                'serviceName': self.object_name('service'),
                'selector': {
                    'matchLabels': {
                        'app': self.object_name('pod-label-app'),
                    }
                },
                'template': {
                    'metadata': {
                        'labels': {
                            'app': self.object_name('pod-label-app'),
                        }
                    },
                    'spec': {
                        'serviceAccountName': ValueData(value=self.object_name('service-account'),
                                                        enabled=self.object_name('service-account') is not None),
                        'initContainers': [{
                            'name': 'init-chown-data',
                            'image': self.option_get('container.init-chown-data'),
                            'imagePullPolicy': 'Always',
                            'command': ['chown', '-R', '65534:65534', '/data'],
                            'volumeMounts': [{
                                'name': 'prometheus-data-volume',
                                'mountPath': '/data'
                            }]
                        }],
                        'containers': [{
                            'name': 'prometheus-server',
                            'image': self.option_get('container.prometheus'),
                            'args': ['--config.file=/etc/prometheus/prometheus.yml',
                                     '--storage.tsdb.path=/data'],
                            'volumeMounts': [{
                                'name': 'prometheus-config-volume',
                                'mountPath': '/etc/prometheus'
                            },
                            {
                                'name': 'prometheus-data-volume',
                                'mountPath': '/data'
                            }],
                            'ports': [{
                                'containerPort': 9090
                            }],
                            'readinessProbe': {
                                'httpGet': {
                                    'path': '/-/ready',
                                    'port': 9090
                                },
                                'initialDelaySeconds': 30,
                                'timeoutSeconds': 30
                            },
                            'livenessProbe': {
                                'httpGet': {
                                    'path': '/-/healthy',
                                    'port': 9090
                                },
                                'initialDelaySeconds': 30,
                                'timeoutSeconds': 30
                            },
                            'resources': ValueData(value=self.option_get('kubernetes.resources.statefulset'),
                                                   disabled_if_none=True),
                        }],
                        'volumes': [
                            {
                                'name': 'prometheus-config-volume',
                                'configMap': {
                                    'name': self.object_name('config'),
                                }
                            },
                            KDataHelper_Volume.info(base_value={
                                'name': 'prometheus-data-volume',
                            }, value=self.option_get('kubernetes.volumes.data')),
                        ]
                    }
                }
            }
        }, name=self.BUILDITEM_STATEFULSET, source=self.SOURCE_NAME, instance=self.basename())]
        return ret



from .builder import (
    PrometheusBuilder,
)
from .option import (
    PrometheusOptions,
)
from .configfile import (
    PrometheusConfigFileOptions,
    PrometheusConfigFile,
)
from .configfileext import (
    PrometheusConfigFileExt_Kubernetes,
)

__version__ = "0.7.8"

__all__ = [
    'PrometheusBuilder',
    'PrometheusOptions',
    'PrometheusConfigFileOptions',
    'PrometheusConfigFile',
    'PrometheusConfigFileExt_Kubernetes',
]



from typing import Mapping

from kubragen.configfile import ConfigFile
from kubragen.kdatahelper import KDataHelper_Volume
from kubragen.option import OptionDef, OptionDefFormat
from kubragen.options import Options


class PrometheusOptions(Options):
    """
    Options for the Prometheus builder.

    .. list-table::
        :header-rows: 1

        * - option
          - description
          - allowed types
          - default value
        * - basename
          - object names prefix
          - str
          - ```prometheus```
        * - namespace
          - namespace
          - str
          - ```prometheus```
        * - config |rarr| prometheus_config
          - prometheus.yml file
          - str, :class:`kubragen.configfile.ConfigFile`
          - :class:`kg_prometheus.PrometheusConfigFile`
        * - config |rarr| service_port
          - service port
          - int
          - ```9090```
        * - config |rarr| authorization |rarr| serviceaccount_create
          - whether to create a service account
          - bool
          - ```True```
        * - config |rarr| authorization |rarr| serviceaccount_use
          - service account to use if not creating
          - str
          -
        * - config |rarr| authorization |rarr| roles_create
          - whether create roles
          - bool
          - ```True```
        * - config |rarr| authorization |rarr| roles_bind
          - whether to bind roles to service account
          - bool
          - ```True```
        * - container |rarr| init-chown-data
          - init-chown-data container image
          - str
          - ```debian:<version>```
        * - container |rarr| prometheus
          - prometheus container image
          - str
          - ```prom/prometheus:<version>```
        * - kubernetes |rarr| volumes |rarr| data
          - Kubernetes data volume
          - dict, :class:`KData_Value`, :class:`KData_ConfigMap`, :class:`KData_Secret`
          -
        * - kubernetes |rarr| resources |rarr| statefulset
          - Kubernetes StatefulSet resources
          - dict
          -
    """
    def define_options(self):
        """
        Declare the options for the Prometheus builder.

        :return: The supported options
        """
        return {
            'basename': OptionDef(required=True, default_value='prometheus', allowed_types=[str]),
            'namespace': OptionDef(required=True, default_value='prometheus', allowed_types=[str]),
            'config': {
                'prometheus_config': OptionDef(required=True, allowed_types=[str, ConfigFile]),
                'service_port': OptionDef(required=True, default_value=9090, allowed_types=[int]),
                'authorization': {
                    'serviceaccount_create': OptionDef(required=True, default_value=True, allowed_types=[bool]),
                    'serviceaccount_use': OptionDef(allowed_types=[str]),
                    'roles_create': OptionDef(required=True, default_value=True, allowed_types=[bool]),
                    'roles_bind': OptionDef(required=True, default_value=True, allowed_types=[bool]),
                },
            },
            'container': {
                'init-chown-data': OptionDef(required=True, default_value='debian:9', allowed_types=[str]),
                'prometheus': OptionDef(required=True, default_value='prom/prometheus:v2.21.0', allowed_types=[str]),
            },
            'kubernetes': {
                'volumes': {
                    'data': OptionDef(required=True, format=OptionDefFormat.KDATA_VOLUME,
                                      allowed_types=[Mapping, *KDataHelper_Volume.allowed_kdata()]),
                },
                'resources': {
                    'statefulset': OptionDef(allowed_types=[Mapping]),
                }
            },
        }



from kubragen.configfile import ConfigFileExtension, ConfigFile, ConfigFileExtensionData
from kubragen.data import ValueData
from kubragen.options import OptionGetter


class PrometheusConfigFileExt_Kubernetes(ConfigFileExtension):
    """
    Prometheus configuration extension for Kubernetes scrape config.

    Based on `prometheus/prometheus <https://github.com/prometheus/prometheus/blob/master/documentation/examples/prometheus-kubernetes.yml>`_.
    Based on `Driving Manual or Automatic: Docker Hub vs. Helm for Deploying Prometheus on Kubernetes <https://logz.io/blog/manual-automatic-kubernetes-prometheus-docker-hub-helm/>`_.
    """
    insecure_skip_verify: bool
    scrape_cadvisor: bool

    def __init__(self, insecure_skip_verify: bool = False, scrape_cadvisor: bool = True):
        """
        :param insecure_skip_verify: set insecure_skip_verify parameter for kubernetes tls_config
        :param scrape_cadvisor: whether to scrape cadvisor (not available in all kubernetes installations)
        """
        self.insecure_skip_verify = insecure_skip_verify
        self.scrape_cadvisor = scrape_cadvisor

    def process(self, configfile: ConfigFile, data: ConfigFileExtensionData, options: OptionGetter) -> None:
        if 'scrape_configs' not in data.data:
            data.data['scrape_configs'] = []
        data.data['scrape_configs'].extend([{
            'job_name': 'kubernetes-apiservers',
            'kubernetes_sd_configs': [{
                'role': 'endpoints'
            }],
            'scheme': 'https',
            'tls_config': {
                'ca_file': '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt',
                'insecure_skip_verify': self.insecure_skip_verify,
            },
            'bearer_token_file': '/var/run/secrets/kubernetes.io/serviceaccount/token',
            'relabel_configs': [{
                'source_labels': ['__meta_kubernetes_namespace',
                                  '__meta_kubernetes_service_name',
                                  '__meta_kubernetes_endpoint_port_name'],
                'action': 'keep',
                'regex': 'default;kubernetes;https'
            }]
        },
        {
            'job_name': 'kubernetes-nodes',
            'scheme': 'https',
            'tls_config': {
                'ca_file': '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt',
                'insecure_skip_verify': self.insecure_skip_verify,
            },
            'bearer_token_file': '/var/run/secrets/kubernetes.io/serviceaccount/token',
            'kubernetes_sd_configs': [{
                'role': 'node'
            }],
            'relabel_configs': [
                {
                    'action': 'labelmap',
                    'regex': '__meta_kubernetes_node_label_(.+)'
                },
            ]
        },
        {
            'job_name': 'kubernetes-pods',
            'kubernetes_sd_configs': [{
                'role': 'pod'
            }],
            'relabel_configs': [{
                'source_labels': ['__meta_kubernetes_pod_annotation_prometheus_io_scrape'],
                'action': 'keep',
                'regex': True
            },
            {
                'source_labels': ['__meta_kubernetes_pod_annotation_prometheus_io_path'],
                'action': 'replace',
                'target_label': '__metrics_path__',
                'regex': '(.+)'
            },
            {
                'source_labels': ['__address__',
                                  '__meta_kubernetes_pod_annotation_prometheus_io_port'],
                'action': 'replace',
                'regex': '([^:]+)(?::\\d+)?;(\\d+)',
                'replacement': '$1:$2',
                'target_label': '__address__'
            },
            {
                'action': 'labelmap',
                'regex': '__meta_kubernetes_pod_label_(.+)'
            },
            {
                'source_labels': ['__meta_kubernetes_namespace'],
                'action': 'replace',
                'target_label': 'kubernetes_namespace'
            },
            {
                'source_labels': ['__meta_kubernetes_pod_name'],
                'action': 'replace',
                'target_label': 'kubernetes_pod_name'
            }]
        },
        ValueData({
            'job_name': 'kubernetes-cadvisor',
            'scheme': 'https',
            'tls_config': {
                'ca_file': '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt',
                'insecure_skip_verify': self.insecure_skip_verify,
            },
            'bearer_token_file': '/var/run/secrets/kubernetes.io/serviceaccount/token',
            'kubernetes_sd_configs': [{
                'role': 'node',
            }],
            'relabel_configs': [
            {
                'action': 'labelmap',
                'regex': '__meta_kubernetes_node_label_(.+)'
            },
            {
                'target_label': '__address__',
                'replacement': 'kubernetes.default.svc:443'
            },
            {
                'source_labels': ['__meta_kubernetes_node_name'],
                'regex': '(.+)',
                'target_label': '__metrics_path__',
                'replacement': '/api/v1/nodes/${1}/proxy/metrics/cadvisor'
            }]
        }, enabled=self.scrape_cadvisor),
        {
            'job_name': 'kubernetes-service-endpoints',
            'kubernetes_sd_configs': [{
                'role': 'endpoints'
            }],
            'relabel_configs': [{
                'source_labels': ['__meta_kubernetes_service_annotation_prometheus_io_scrape'],
                'action': 'keep',
                'regex': True
            },
            {
                'source_labels': ['__meta_kubernetes_service_annotation_prometheus_io_scheme'],
                'action': 'replace',
                'target_label': '__scheme__',
                'regex': '(https?)'
            },
            {
                'source_labels': ['__meta_kubernetes_service_annotation_prometheus_io_path'],
                'action': 'replace',
                'target_label': '__metrics_path__',
                'regex': '(.+)'
            },
            {
                'source_labels': ['__address__',
                                  '__meta_kubernetes_service_annotation_prometheus_io_port'],
                'action': 'replace',
                'target_label': '__address__',
                'regex': '([^:]+)(?::\\d+)?;(\\d+)',
                'replacement': '$1:$2'
            },
            {
                'action': 'labelmap',
                'regex': '__meta_kubernetes_service_label_(.+)'
            },
            {
                'source_labels': ['__meta_kubernetes_namespace'],
                'action': 'replace',
                'target_label': 'kubernetes_namespace'
            },
            {
                'source_labels': ['__meta_kubernetes_service_name'],
                'action': 'replace',
                'target_label': 'kubernetes_name'
            }],
        }])
